# Copyright (c) 2004 Simplistix Ltd
#
# This Software is released under the MIT License:
# http://www.opensource.org/licenses/mit-license.html
# See license.txt for more details.

ManageUsersPermission = 'Manage users'
ViewManagementPermission = 'View management screens'

from AccessControl import ClassSecurityInfo, getSecurityManager
from AccessControl.User import UserFolder, SimpleUser, DEFAULTMAXLISTUSERS, _noroles
from BTrees.OOBTree import OOBTree
from DateTime import DateTime
from OFS.ObjectManager import ObjectManager, REPLACEABLE
from Globals import InitializeClass
from Globals import DTMLFile
from re import compile
from urllib import urlopen, quote as url_quote
from ZPublisher.HTTPRequest import HTTPRequest
from Shared.DC.ZRDB.Results import Results
from xml.dom.minidom import parseString
from zLOG import LOG, ERROR

import random
import sys

try:
    jn = random.randint(0,2L**160)
    def getRandom(self):
        return hex(random.randint(0,2**160))[2:-1]
except:
    # python < 2.3
    jn = random.random()
    def getRandom(self):
        return str(random.random())[2:]
    
class CASError(Exception):
    "An error originating from the CAS Server"
    pass

# paranoid scrambling
random.jumpahead(jn)

# the default store of session information
import threading
lock = threading.Lock()
session2name = OOBTree()
time2session = OOBTree()

# pick the username out of the CAS response
username_re = compile('<cas:user>([^<]+)</cas:user>')

def addCASUserFolder(self,REQUEST=None):
    """Add a CASUserFolder to a container as acl_users"""
    ob = CASUserFolder()
    self._setObject('acl_users', ob)
    if REQUEST is not None:
        return self.manage_main(self,REQUEST)

class CASUserFolder(ObjectManager,UserFolder):

    meta_type = "CAS User Folder"

    security = ClassSecurityInfo()

    # default values
    cas_login_url = ''
    cas_validate_url = ''
    session_cookie_path = '/'
    session_cookie_name = 'cuf_session_cookie'
    session_cookie_rotate = 0
    session_cookie_secure = 0
    session_cookie_expiry = 0
    session_timeout = 10

    security = ClassSecurityInfo()

    # modified ZMI
    manage_options=(
        ({'label':'Users', 'action':'manage_main_users',
         'help':('OFSP','User-Folder_Contents.stx')},)
        + ObjectManager.manage_options[0:1]
        + UserFolder.manage_options[1:]
        )

    security.declareProtected(ViewManagementPermission,'manage_main')
    manage_main = ObjectManager.manage_main
    
    security.declareProtected(ManageUsersPermission,'manage_main_users')
    manage_main_users = UserFolder.manage_main

    # modified DTML files
    _userFolderProperties = DTMLFile('dtml/userFolderProps', globals())
    
    _add_User=DTMLFile('dtml/addUser', globals())
    
    _editUser=DTMLFile('dtml/editUser', globals())

    # editing our properties
    def manage_setUserFolderProperties(self, encrypt_passwords=0,
                                       update_passwords=0,
                                       maxlistusers=DEFAULTMAXLISTUSERS,
                                       cas_login_url=None,
                                       cas_validate_url=None,
                                       session_cookie_path = '/',
                                       session_cookie_name = 'cuf_session_cookie',
                                       session_cookie_rotate = 0,
                                       session_cookie_secure = 0,
                                       session_cookie_expiry = 0,
                                       session_timeout = 0,
                                       REQUEST=None):
        """
        Sets the properties of the user folder.
        """
        self.cas_login_url = cas_login_url
        self.cas_validate_url = cas_validate_url
        self.session_cookie_path = session_cookie_path
        self.session_cookie_name = session_cookie_name
        self.session_cookie_rotate = session_cookie_rotate
        self.session_cookie_secure = session_cookie_secure
        self.session_cookie_expiry = session_cookie_expiry
        self.session_timeout = session_timeout
        return UserFolder.manage_setUserFolderProperties(self, encrypt_passwords,
                                                         update_passwords,
                                                         maxlistusers,
                                                         REQUEST)

    # overridable methods
    def generateSession(self):
        """supply a string that is to be used as the session
        identifier. Should return a sequence of random characters
        suitable for storing as a cookie value.

        This method MUST be quick since it is called at least once
        for every request
        """
        raise NotImplementedError

    # override with whatever works on our platform
    generateSession = getRandom
        
    generateSession__replaceable__ = REPLACEABLE

    def doLogout(self,username,session):
        """logs a user out of one or all active sessions at
        the implementors discretion.
        
        username is the user to be logged out.
        session is the particular session id to log out.
        (session may be None)
        """
        pass
    doLogout__replaceable__ = REPLACEABLE

    def getUserName(self,session):
        """retrieves the stored username for a given session id.

        session is the session id

        should return either None, the username as a string or a
        Shared.DC.ZRDB.Results.Results object.
        """
        lock.acquire()
        username = session2name.get(session,None)
        lock.release()
        return username
    getUserName__replaceable__ = REPLACEABLE

    def deleteSession(self,session):
        """deletes the specified session

        session is the session id.
        """
        global session2name
        global time2session        
        lock.acquire()
        del session2name[session]
        lock.release()
    deleteSession__replaceable__ = REPLACEABLE

    def expireSessions(self,expiretime):
        """expires all session with an expiry time set
        before the supplied expiry time.

        expiretime is a floating point number
        """
        global session2name
        global time2session        
        lock.acquire()
        for time,sessions in time2session.items(0,expiretime):
            for session in sessions:
                try:                    
                    del session2name[session]
                except KeyError:
                    # it's already gone
                    pass
            del time2session[time]
        lock.release()
    expireSessions__replaceable__ = REPLACEABLE

    def storeSession(self,username,expiretime,session):
        """stores the supplied username, expiretime and session
        in an appropriate data store.

        username is a string
        expiretime is a floating point number
        session is the session id as a string
        """
        global session2name
        global time2session        
        lock.acquire()
        session2name[session] = username
        if not time2session.has_key(expiretime):
            time2session[expiretime] = []
        time2session[expiretime].append(session)
        lock.release()
    storeSession__replaceable__ = REPLACEABLE

    def checkProxies(self,proxies):
        """check whether the specified proxies are allowed.
        Should return a true value if they are and a false value
        if they are not.
        
        proxies is a list of strings.
        Each string is the pgtUrl of the proxying service.
        If proxies contains more than one entry, the order will
        be from the first proxying application to the last
        proxying application.
        """
        return 0
    checkProxies__replaceable__ = REPLACEABLE
    
    # helper function
    def __computeCameFrom(self,request):
        # compute where we're coming from
        came_from = request.get('URL', '')
        query = request.get('QUERY_STRING')
        if query:
            nq=[]
            # strip out any tickets, this will keep the
            # 'service' we send for CAS validation the
            # same as the service got the ticket with            
            for bit in query.split('&'):
                k = bit.split('=')[:1]
                if not bit or k[0]=='ticket':
                    continue
                nq.append(bit)
            if nq:
                came_from += ('?'+'&'.join(nq))
        return url_quote(came_from)
        
    security.declarePublic('logout')
    def logout(self,redirect=None):
        """End the local non-CAS session.
           On any page that calls this, or is specified in the redirect parameter,
           you may want to put a link to end the cas session too."""
        request = self.REQUEST
        session = getattr(request,'_cas_session',None)
        response = request.RESPONSE
        response.expireCookie(self.session_cookie_name,
                              path=self.session_cookie_path)

        
        self.doLogout(username=getSecurityManager().getUser().getId(),
                      session=session)
        
        if redirect:
            return response.redirect(redirect)
        
    
    def validate(self, request, auth='', roles=_noroles):
        """
        this method performs identification, authentication, and
        authorization
        v is the object (value) we're validating access to
        n is the name used to access the object
        a is the object the object was accessed through
        c is the physical container of the object
        t is the current time as DateTime
        tt is the current time as seconds since the epoch
        username is the name of the authenticated user
        user is the user object for the authenticated user

        We allow the publishing machinery to defer to higher-level user
        folders or to raise an unauthorized by returning None from this
        method.
        """

        v = request['PUBLISHED'] # the published object
        a, c, n, v = self._getobcontext(v, request)
        t = DateTime()
        tt = int(t.timeTime())

        # get ticket and ALWAYS delete so as not to confuse things.
        # You may wish to comment the section between the DDD's out when
        # debugging
        
        # DDD
        ticket = request.form.get('ticket')
        if ticket is not None:
            del request.form['ticket']
        # DDD
            
        # bail if we're not a plain web request and defer to a higher userfolder
        if (request.__class__ is not HTTPRequest
            or not request['REQUEST_METHOD'] in ('HEAD', 'GET', 'PUT', 'POST')
            or request.environ.has_key('WEBDAV_SOURCE_PORT')):
            return None

        # request-based overrides to turn off CAS auth or re-enable it from a particular
        # browser.
        # Handy if the CAS auth user can't do something you want to, but you have basic auth
        # details mapped to a user which can. Could be necessary for emergency user ;-)
        if request.get('enable_cas_auth'):
            # delete any set cookie
            request.RESPONSE.expireCookie('disable_cas_auth',path='/')
            # modify the request, so the code below will work this request
            request.set('disable_cas_auth',None)
        if request.get('disable_cas_auth'):
            # set a cookie, we'll delete it later
            request.RESPONSE.setCookie('disable_cas_auth','1',path='/')
            # bail and defer to a higher userfolder
            return None
    
        username = None
        
        # opportunistically time out old sessions
        self.expireSessions(expiretime=tt)
        
        # check for an existing "local" cookie-based session
        current_session = request.cookies.get(self.session_cookie_name)
        request._cas_session = current_session

        # Delete CAS session from request. You may wish to comment
        # the section between the DDD's out when debugging

        # DDD
        try:
            del request.cookies[self.session_cookie_name]
        except:
            pass
        try:
            # we delete the whole HTP_COOKIE header for speed
            # no-one should be using it anyway
            del request.environ['HTTP_COOKIE']
        except:
            pass
        # DDD
        
        if current_session:
            # get the username
            username = self.getUserName(session=current_session)
            # support Z SQL Method implementations of getUserName
            if isinstance(username,Results):
                if username:
                    username = username[0].username
                else:
                    username = None
            if username is None:
                # we have a bogus cookie, so delete it and logout the
                # current user from the local session
                self.logout()
            elif self.session_cookie_rotate:
                # we've used that session, so discard it
                self.deleteSession(session=current_session)
            
        # otherwise check if we have a CAS ticket
        elif ticket and self.cas_validate_url:
            # validate the ticket
            xml = urlopen(self.cas_validate_url+'?service='+self.__computeCameFrom(request)+'&ticket='+ticket).read()
            # parse the xml, do nothing if it fails to parse
            try:
                domdoc = parseString(xml)
            except:
                # not valid XML, make a note and move on.
                raise CASError, 'Invalid XML returned:\n' + xml
            else:
                # check we have appropriate CAS xml
                domdoc.normalize()
                de = domdoc.documentElement
                if de.tagName==u'cas:serviceResponse':
                    asl = de.getElementsByTagName('cas:authenticationSuccess')
                    if len(asl)==1:
                        as = asl[0]
                        for c in as.childNodes:
                            if c.nodeType==c.ELEMENT_NODE:
                                if c.tagName==u'cas:user':
                                    cn = c.childNodes
                                    # paranoid check we only have one child
                                    assert len(cn)==1
                                    # the following will break if username
                                    # contains any non-ascii characters
                                    username=str(cn[0].nodeValue)
                                elif c.tagName==u'cas:proxies':
                                    proxies = []
                                    for p in c.childNodes:
                                        if p.nodeType==p.ELEMENT_NODE:
                                            # paranoid check we have a proxy tag
                                            assert p.tagName=='cas:proxy'
                                            # paranoid check it only has one child
                                            pcn = p.childNodes
                                            assert len(pcn)==1
                                            # the following will break if pgtUrl
                                            # contains any non-ascii characters                                            
                                            proxies.insert(0,str(pcn[0].nodeValue))
                                    if not self.checkProxies(proxies):
                                        raise CASError, 'Disallowed proxy chain:\n'+', '.join(proxies)
                                else:
                                    raise CASError,'Unkown authentication success returned:\n' + xml
                    else:
                        afl = de.getElementsByTagName('cas:authenticationFailure')
                        if len(afl)==1:
                            af = afl[0]
                            cn = af.childNodes
                            # paranoid check we only have one child
                            assert len(cn)==1
                            # the following will break if error code
                            # contains any non-ascii characters
                            code = str(af.getAttribute('code'))
                            # the following will break if error message
                            # contains any non-ascii characters
                            message = str(cn[0].nodeValue)
                            raise CASError,code+':'+message
                        else:
                            raise CASError,'Neither cas:authenticationSuccess or cas:authenticationFailure found in XML:\n' + xml
                else:
                    raise CASError, 'cas:serviceResponse not found in XML:\n' + xml

        # if we still have no username, 
        if username is None:
            
            # is anonymous authorized to view the resource?
            if self.authorize(self._nobody, a,c,n,v,roles):
                # yes, so stay anonymous
                return self._nobody.__of__(self)
            
            # anonymous isn't authorized, and we have no local session, so
            # we had better go off to the CAS server and try and authenticate there
            if not self.cas_login_url:
                # we haven't been configured yet, so pretend we aren't here
                return None
            raise 'Redirect', self.cas_login_url + '?service=' + self.__computeCameFrom(request)

        # we have a username, so store a new local session if needed
        if self.session_cookie_rotate or not current_session:
            
            session = self.generateSession()
            te = tt + self.session_timeout*60
            self.storeSession(username=username,
                              expiretime=te,
                              session=session)
                
            # cookie attributes
            cookie = {
                'path':self.session_cookie_path,
                'secure':self.session_cookie_secure,
                }
        
            # cookie expiry time
            if self.session_cookie_expiry:
                cookie['expires']=(
                    t + self.session_timeout/1440.0
                    ).toZone('GMT').rfc822()

            request.RESPONSE.setCookie(self.session_cookie_name,
                                       session,
                                       **cookie)

        # get the user object, if we have one
        # or create a new temp. user if one doesn't exist
        user = self.getUser(username)
        if user is None:
            user = SimpleUser(username,'',(),'').__of__(self)

        # is the user we now have authorized?
        if self.authorize(user, a, c, n, v, roles):
            # yes, so return it
            return user.__of__(self)

        # no, so defer to a higher userfolder
        return None
    
    # other methods
    def manage_afterAdd(self,item,container):
        ObjectManager.manage_afterAdd(self,item,container)
        UserFolder.manage_afterAdd(self,item,container)

    def manage_beforeDelete(self,item,container):
        UserFolder.manage_beforeDelete(self,item,container)    
        ObjectManager.manage_beforeDelete(self,item,container)

InitializeClass(CASUserFolder)


